//--------------------------------------------------------------------------------------------
/// \file	main.cpp
/// \author	Rob Bateman
/// \brief	Probably the simplest maya exporter code possible. This will work on both linux
/// 		and win32, makes sure that your VisualC++ options has set the correct paths to
/// 		the maya includes and libraries. When compiling on linux, make sure you have
/// 		set the correct paths contained within the makefile that is included.
///
/// 		When compiling under linux, you will need to have gtk2 installed (the file dialog
/// 		code uses it. It would be fairly trivial to strip that out however).
//--------------------------------------------------------------------------------------------

#include <maya/MItDependencyNodes.h>
#include <maya/MItDag.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MFnWeightGeometryFilter.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MPlugArray.h>
#include <maya/MPlug.h>
#include <maya/MDagPathArray.h>
#include <maya/MDagPath.h>
#include <maya/MFileIO.h>
#include <maya/MLibrary.h>
#include <maya/MFnDagNode.h>
#include <maya/MFnSkinCluster.h>

#include <fstream>
#include <map>
#include <string>
#include <vector>
#include <algorithm>

using namespace std;

// generic linux(gtk2) & win32 file dialog code
#include "FileDialog.h"

// link to some fairly funky maya libs. NOTE: The Image library only became part of maya
// from version 4.01 onwards. It is not available in version 4.0
//
#ifdef WIN32
	#pragma comment(lib,"Foundation.lib")
	#pragma comment(lib,"OpenMaya.lib")
	#pragma comment(lib,"OpenMayaFx.lib")
	#pragma comment(lib,"Image.lib")
	#pragma comment(lib,"OpenMayaAnim.lib")
#endif

//--------------------------------------------------------------------------------------------
//
//
void OutputMaterial(MObject& obj) {
	MFnDependencyNode fn(obj);
	cout << "mat:\t" << fn.name().asChar() << endl;
}
void OutputShape(MObject& obj) {
	MFnDagNode fn(obj);
	cout << "shape:\t" << fn.name().asChar() << endl;
}
void OutputTexture(MObject& obj) {
	MFnDependencyNode fn(obj);
	cout << "tex:\t" << fn.name().asChar() << endl;
}
void OutputTransform(MObject& obj) {
	MFnDagNode fn(obj);
	cout << "xform:\t" << fn.name().asChar() << endl;
}

//--------------------------------------------------------------------------------------------

// global map of texture references
map<string,MObject> Textures;
// global map of material references
map<string,MObject> Materials;
// global map of material references
map<string,MObject> Shapes;
// global map of material references
map<string,MObject> Transforms;


void CleanReferences() {   
	Textures.clear();
	Materials.clear();
	Shapes.clear();
	Transforms.clear(); 
}

void AddTexture(MObject& obj) {
	MFnDependencyNode fn(obj);

	// get name of texture node 
	string name = fn.name().asChar();

	// if we can't find it in the map...
	if( Textures.find( name ) == Textures.end() ) {

		// insert the object into the map 
		Textures.insert( make_pair(name,obj) ); 

	}
}

void AddMaterial(MObject& obj) 
{
	MFnDependencyNode fn(obj);

	// get name of texture node 
	string name = fn.name().asChar();

	// if we can't find it in the map...
	if( Materials.find( name ) == Materials.end() ) {

		// insert the object into the map 
		Materials.insert( make_pair(name,obj) ); 
	}
}

void AddObject(MObject& obj) 
{
	MFnDependencyNode fn(obj);

	// get name of texture node 
	string name = fn.name().asChar();

	// if we can't find it in the map...
	if( Shapes.find( name ) == Shapes.end() ) {

		// insert the object into the map 
		Shapes.insert( make_pair(name,obj) ); 
	}
}

void AddTransform(MObject& obj) 
{
	MFnDependencyNode fn(obj);

	// get name of texture node 
	string name = fn.name().asChar();

	// if we can't find it in the map...
	if( Transforms.find( name ) == Transforms.end() ) {

		// insert the object into the map 
		Transforms.insert( make_pair(name,obj) ); 

	}
}

void  RecurseUp(MObject& bone)
{ 
	MFnDagNode fn(bone);

	// add the transform into the export list
	AddTransform( fn.object() ); 

	// loop through each parent of this transform. (We may have
	// more than one parent if we have an instance)
	for(int i=0;i<fn.parentCount();++i) {
		// add parent node into transform list
		RecurseUp( fn.parent(i) );
	}
}


void ReferenceSkinClusters()
{ 
	MItDependencyNodes it(MFn::kSkinClusterFilter);

	// loop through each skin cluster
	for( ; !it.isDone(); it.next() ) {

		MStatus stat;

		// attach a skin cluster function set to
		// access the data
		MFnSkinCluster fn(it.item());
		MDagPathArray infs;

		// get an array of influence objects
		fn.influenceObjects(infs,&stat);

		// add each bone in skin cluster to transform list
		for(int i=0;i<infs.length();++i) { 

			RecurseUp( infs[i].node() ); 

		}
	}
}

void ReferenceMaterial(MObject& material) 
{
	MFnDependencyNode fn(material);

	// add the material to our export list 
	AddMaterial(material);

	// get an array of all connections to this node
	MPlugArray plugs;
	fn.getConnections(plugs);

	// loop through all connected attributes on this node
	for(int i=0;i<plugs.length();++i) {

		MPlugArray conns;

		// get the connections to this attribute 
		plugs[i].connectedTo(conns,true,true);

		// loop through each connection looking for a
		// file texture node.
		for(int j=0;j<conns.length();++j) {

			MObject obj = conns[j].node();

			if(obj.apiType()==MFn::kFileTexture) { 

				// add texture to export list
				AddTexture(obj);

			}
		}
	}
}


void ReferenceObject(MObject& shape)
{ 

	MFnDagNode fn(shape);

	if( fn.isIntermediateObject() ) return ;

	AddObject( fn.object() ); 

	// make sure we reference all of the objects 
	// parent transforms
	for(int i=0;i<fn.parentCount();++i) {

		RecurseUp( fn.parent(i) );
	}

	// get an array of all connections to this node
	MPlugArray plugs;
	fn.getConnections(plugs);

	// loop through all connected attributes on this node
	for(int i=0;i<plugs.length();++i) {

		MPlugArray conns;

		// get the connections to this attribute 
		plugs[i].connectedTo(conns,true,true);

		// loop through each connection looking for a
		// material node.
		for(int j=0;j<conns.length();++j) {

			MObject obj = conns[j].node();

			if( obj.hasFn(MFn::kShadingEngine) ) { 

				MFnDependencyNode fnShade(obj);

				MPlug mat_plg = fnShade.findPlug("surfaceShader");

				MPlugArray material_conns;

				// get the connections to this attribute 
				mat_plg.connectedTo(material_conns,true,true);

				for(int k=0;k<material_conns.length();++k) {

					MObject mat = material_conns[k].node();

					if( mat.hasFn(MFn::kLambert) ) {	
						// add texture to export list
						ReferenceMaterial(mat);
					}

				}

			}

			if( shape.hasFn(MFn::kParticle) ) {

				if( obj.hasFn(MFn::kField) ) { 

					// add dynamics to export list
					RecurseUp(obj);

				}

				if( obj.hasFn(MFn::kEmitter) ) { 

					// add emitter to export list
					RecurseUp(obj);
				}
			}
		}
	}
}

void ReferenceJointClusters()
{ 

	MItDependencyNodes it(MFn::kJointCluster);

	// loop through each skin cluster
	for( ; !it.isDone(); it.next() ) {



		// attach a skin cluster function set to
		// access the data
		MFnWeightGeometryFilter fn(it.item());

		MObject joint;

		// get a plug to the matrix attribute
		MPlug pJointPlug = fn.findPlug("matrix");
		MPlugArray conns;

		// add each bone in skin cluster to transform list
		if (pJointPlug.connectedTo(conns,true,false)) {

			RecurseUp( conns[0].node() ); 

		}
	}
}


void ParseScene()
{ 
	CleanReferences();

	MItDag it;

	// loop through each object in the scene
	for( ; !it.isDone(); it.next() ) {


		switch(it.item().apiType()) {

		// output any 'interesting' shapes we find
		case MFn::kMesh:
		case MFn::kNurbsCurve:
		case MFn::kNurbsSurface:
		case MFn::kRenderSphere:
		case MFn::kRenderBox:
		case MFn::kRenderCone:
		case MFn::kLocator:
		case MFn::kPointLight:
		case MFn::kAmbientLight:
		case MFn::kDirectionalLight:
		case MFn::kSpotLight:
		case MFn::kParticle:
		case MFn::kLattice:
		case MFn::kBaseLattice:
			ReferenceObject( it.item() );
			break;

		default : break;

		} 
	}

	// make sure we include bones used in skinning 
	ReferenceSkinClusters();
	ReferenceJointClusters();
}


void Export()
{ 



	ParseScene();

	{ // output all textures

		map<string,MObject>::iterator it;

		it = Textures.begin();

		// loop through each object in the scene
		for( ; it != Textures.end(); ++it ) {

			OutputTexture(it->second);

		}

	}

	{ // output all materials

		map<string,MObject>::iterator it;

		it = Materials.begin();

		// loop through each object in the scene
		for( ; it != Materials.end(); ++it ) {

			OutputMaterial(it->second);
		}
	}

	{ // output all shapes

		map<string,MObject>::iterator it;

		it = Shapes.begin();

		// loop through each object in the scene
		for( ; it != Shapes.end(); ++it ) {

			OutputShape(it->second);
		}
	}

	{ // output all transforms
		map<string,MObject>::iterator it;

		it = Transforms.begin();

		// loop through each object in the scene
		for( ; it != Transforms.end(); ++it ) {

			OutputTransform(it->second);
		}
	}

	// All that remains is to output the following:
	//
	// Skin & Joint Clusters
	// Blend Shapes
	// Free Form Deformations
	// Animation Data
	// 

	// clean all temp data
	CleanReferences();
}







#ifdef WIN32
int main(int argc,char** argv)
{
#else // hack for linux & gtk

// gtk needs the command line args for initialisation
int   g_argc=0;
char** g_argv=0;

int main(int argc,char** argv)
{
	g_argc=argc;
	g_argv=argv;
#endif
	char InFileName[512]={0};
	char OutFileName[512]={0};

	// both an infile and outfile specified as command line args
	if (argc == 3 )
	{
		strcpy(InFileName,argv[1]);
		strcpy(OutFileName,argv[2]);
	}
	else

	// if only one argument was specified, I'll assume that it is a maya binary
	// or ascii file
	if (argc == 2 )
	{
		// got infile from command line args
		strcpy(InFileName,argv[1]);

		// ask for an outfile
		if(!SaveFileDialog(OutFileName)) {
			return EXIT_FAILURE;
		}
	}
	else

	// no args specified so pop up a couple fo file dialogs
	if (argc == 1 )
	{
		// get in file name
		if(!OpenFileDialog(InFileName)) {
			return EXIT_FAILURE;
		}

		// get out file name
		if(!SaveFileDialog(OutFileName)) {
			return EXIT_FAILURE;
		}
	}
	else {
		cerr << "[ERROR] Usage\n\tSimpleExporter [infile.mb|infile.ma] [outfile]\n";
		return EXIT_FAILURE;
	}

	// initialise the maya library - This basically starts up Maya
	MLibrary::initialize(argv[0]);


	// open the file that the user specified
	if( MS::kSuccess == MFileIO::open(InFileName,NULL,true) )
	{
		Export();
	}
	else {
		cerr << "[ERROR] Could not open file " << InFileName << endl;
	}

	// close down maya
	MLibrary::cleanup(0);

	return 0;
}
